from typing import List
from .create_action import Action
from .create_reducer import Reducer



def combine_reducers(reducers: List[Reducer]) -> Reducer:
  """
  Combines reducers.
  
  This allows the developer to combine multiple reducers into
  one.

  The resulting reducer function will take the state produced
  by the first reducer, and pass that as the initial value to
  the second reducer. The resulting state is constructed by
  taking the result of one reducer, then passing that as the
  initial value to the next reducer, and so on.

  A common use case for this API is to allow each reducer
  to manage an independent part of the global state, while
  allowing each reducer to operate independently.
  
  Args:
  - reducers: The list of reducers to combine.
  
  Returns:
  - A combined reducer.
  """

  def combined_reducers(state: dict, action: Action) -> Reducer:
    next_state = state
    for reducer in reducers:
      next_state = reducer(next_state, action)
    return next_state

  return combined_reducers
  



from typing import Callable, Dict
from .create_action import Action

type Reducer[State: dict, ActionType: Action] = Callable[[State, ActionType], State]

def create_reducer[S: dict, A: Action](initial_state: dict, handlers: Dict[str, Reducer[dict, Action]]) -> Reducer[dict, Action]:
  """
  create a reducer function for use by the store.
  
  Args:
  - initial_state: The initial state of the reducer.
  - handlers: A dictionary of action types and their corresponding reducer functions.
  
  Returns:
  - A reducer function that can be used with the Redux store.
  """

  def reducer_fn(state: S, action: A) -> S:
    if action.type == "@@INIT":
      return initial_state
    if action["type"] in handlers:
      return handlers[action["type"]](state, action)
    else:
      return state

  return reducer_fn
  


# from typing import Callable
# from .create_action import Action, Payload, ActionCreator

# def create_async_thunk[T: str, P: Payload](t: str, thunk: Callable[[Payload], Action]) -> ActionCreator[str, Payload]:
#   """
#   Returns a function that creates an action object with the given type and payload.
#   Actions should be dispatched to the store using the store.dispatch() method.

#   Args:
#   - t (str): The type of the action.

#   Returns:
#   - action_creator (function): A function that creates an action object with the given type and payload.
#   """
#   def thunk_action_creator(payload: P) -> Action[T, P]:
#     return { "type": t, "payload": payload,  thunk: thunk}
#   return thunk_action_creator


"""
This module defines the Action class and related types for creating actions in pyflux.

Payload: A type alias for a dictionary or None.
Action: A dictionary with a "type" key and a "payload" key.
ActionCreator: A callable that takes a payload and returns an Action.

"""
from typing import Callable, TypeVar

type Payload[PT: dict] = PT | None
type Action[AT: str, AP: Payload] = { "type": AT, "payload": AP}
type ActionCreator[TC: str, PC: Payload] = Callable[[PC],Action[TC, PC]]



def create_action[T: str, P: Payload](t: str) -> ActionCreator[str, Payload]:
  """
  Returns a function that creates an action object with the given type and payload.
  Actions should be dispatched to the store using the store.dispatch() method.

  Args:
  - t (str): The type of the action.

  Returns:
  - action_creator (function): A function that creates an action object with the given type and payload.
  """
  def action_creator(payload: P) -> Action[T, P]:
    return { "type": t, "payload": payload }
  return action_creator



from .create_reducer import create_reducer, Reducer
from typing import Dict

def create_slice(reducers:Dict[str, Reducer]={}) -> Reducer:
  def call_reducers(state, reducers):
      """
      Calls each reducer in the given dictionary with the corresponding key in the state object.
      """
      for key, reducer in reducers.items():
          state[key] = reducer(state[key])
      return state
  return call_reducers

  



"""
This module implements a simple state management system inspired by ReduxJS.

The core concept of this system is the store, which holds the state of the application and provides a way to dispatch actions that modify that state. The store is created by passing a reducer function to the createStore function.

The reducer function takes the current state and an action as arguments, and returns a new state based on the action. Actions are plain objects that have a type field and an optional payload field.

This module also provides some utility functions for working with actions and reducers, such as createAction and combineReducers.

Example usage:

  from pyflux import create_action, create_store, create_reducer

  # Define some actions
  increment = create_action('INCREMENT')
  decrement = create_action('DECREMENT')

  # Define a reducer
  def counter(state=0, action):
    if action.type == 'INCREMENT':
      return state + 1
    elif action.type == 'DECREMENT':
      return state - 1
    else:
      return state

  # Combine multiple reducers into one
  rootReducer = create_reducer({
    'increment': counter,
    'decrement': counter,
  })

  # Create the store
  store = create_store(rootReducer)

  # Dispatch some actions
  store.dispatch(increment())
  store.dispatch(decrement())
"""



import asyncio
from threading import Thread
from .create_reducer import create_reducer, Reducer
from .create_action import Action


type StoreBehavior = "greedy" | "lazy"



def create_store(reducer, initial_state=None, store_behavior: StoreBehavior = "greedy"):
  """
  Creates a store object that holds the complete state tree of your app.
  The only way to change the state inside it is to dispatch an action on it.
  """
  state = initial_state
  dispatches = []
  thunks = []




  def get_state():
    if store_behavior=="lazy":
      # for i, thunk in enumerate(thunks):
      #   if not thunk.is_alive():
      #     thunks.pop(i)
      #     dispatches.append(thunk.join())
      for dispatch in dispatches:
        state = reducer(state, dispatch)
      dispatches.clear()
    
    return state

  def dispatch(action: Action):
    # if "thunk" in action:
    #   thunks.append(Thread(target=lambda : asyncio.run(action.thunk(action.payload))).start())
    # if store_behavior=="greedy":
    #   for i, thunk in enumerate(thunks):
    #     if not thunk.is_alive():
    #       thunks.pop(i)
    #       dispatches.append(thunk.join())
      for dispatch in dispatches:
        state = reducer(state, dispatch)
    dispatches.append(action)

  return {
    "get_state": get_state,
    "dispatch": dispatch
  }
